package edu.usc.ict.gordon.copa.eval;

import edu.usc.ict.gordon.util.*;

import java.util.Date;
import java.util.Random;
import java.util.Hashtable;
import java.util.Enumeration;

import java.io.File;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;

import static java.lang.System.out;

/**
 * Tool for determining whether two systems are significantly different in
 * their performance on the COPA datasets. The significance test is based on
 * approximate randomization that involves a stratified shuffling approach in
 * order to build a distribution of differences in performance between the two
 * systems. 
 * <p>
 * The method of approximate randomization is described in:
 * <br>Eric W. Noreen. 1989. Computer-Intensive Methods for Testing Hypotheses: An 
 * Introduction. John Wiley &amp; Sons.
 * <p>
 * A description of how this method was used for MUC-4 data is in:
 * <br>Nancy Chinchor. 1992. The Statistical Significance of the MUC-4 Results. Proceedings of the
 * Fourth Message Understanding Conference (MUC-4).
 * <p>
 * Another implementation of this method can be found at
 * <a href="http://www.cis.upenn.edu/~dbikel/software.html#comparator">this address</a>.
 *
 * @author Cosmin Adrian Bejan
 * @version 1.0
 * @since JDK1.6 | August 2010
 */
public class SignificanceTest {
  /* number of shuffling iterations */
  private static final int NUMBER_OF_SHUFFLES = 9999;
  /* number of categories considered */
  private static final int NUMBER_OF_CATEGORIES = 2;

  /* map for storing gold results: instance id - category */
  private Hashtable<Integer, Integer> mapRGold; 
  /* map for storing system1 results:  instance id - <category, mutable category> */
  private Hashtable<Integer, IntPair> mapRSystem1;
  /* map for storing system2 results:  instance id - <category, mutable category> */  
  private Hashtable<Integer, IntPair> mapRSystem2;
  
  
  /**
   * Initiates the maps for storing the gold and system responses
   *
   */
  public SignificanceTest() {
    mapRGold = new Hashtable<Integer, Integer>();
    mapRSystem1 = new Hashtable<Integer, IntPair>();
    mapRSystem2 = new Hashtable<Integer, IntPair>();
  }
  
  /**
   * Loads into maps all the gold and system responses
   * 
   * @param goldFilePath the file with gold responses
   * @param system1FilePath the file with the responses determined by system1
   * @param system2FilePath the file with the responses determined by system2 
   */
  public void loadResponses(String goldFilePath, String system1FilePath, String system2FilePath) {
    BufferedReader brGold = null;
    BufferedReader brSystem1 = null;
    BufferedReader brSystem2 = null;
    String goldLine = null;
    String system1Line = null;
    String system2Line = null;
    try {
      brGold = new BufferedReader(new FileReader(goldFilePath));
      brSystem1 = new BufferedReader(new FileReader(system1FilePath));
      brSystem2 = new BufferedReader(new FileReader(system2FilePath));
      
      while(true) {
        goldLine = brGold.readLine();
        system1Line = brSystem1.readLine();
        system2Line = brSystem2.readLine();
        if(goldLine == null || system1Line == null || system2Line == null) break;       
        
        updateReponses(new ResponseItem(goldLine.trim(), NUMBER_OF_CATEGORIES), 
                       new ResponseItem(system1Line.trim(), NUMBER_OF_CATEGORIES), 
                       new ResponseItem(system2Line.trim(), NUMBER_OF_CATEGORIES));
      }
      if(goldLine != null || system1Line != null || system2Line != null)
        throw new UnsupportedDataFormatException("Different number of lines in the response files!");
    } catch(IOException ioe) {
      out.println("Error loading response files:"+ioe.getMessage());
      ioe.printStackTrace();
    } finally {
      try {
        if(brGold != null) brGold.close();
        if(brSystem1 != null) brSystem1.close();
        if(brSystem2 != null) brSystem2.close();        
      } catch (IOException ioe) {
        out.println("Error in closing the response file: "+ioe.getMessage());
        ioe.printStackTrace();
      }
    }            
  }
  
  /**
   * Updates the responses into maps
   *
   * @param goldR the gold key associated with one question
   * @param system1R the response determined by system1 for one question
   * @param system2R the response determined by system2 for one question
   */
  private void updateReponses(ResponseItem goldR, ResponseItem system1R, ResponseItem system2R) {
    int goldID = goldR.getID();
    int s1ID = system1R.getID();
    int s2ID = system2R.getID();
    
    if(goldID != s1ID || goldID != s2ID)
      throw new UnsupportedDataFormatException("Incompatible ids gold ["+goldID+"] system1 ["+s1ID+"] system2 ["+s2ID+"]");    
    if(mapRGold.containsKey(goldID))
      throw new UnsupportedDuplicateException("gold id ["+goldID+"] already in the map!");
    if(mapRSystem1.containsKey(s1ID))
      throw new UnsupportedDuplicateException("system1 id ["+s1ID+"] already in the map!");
    if(mapRSystem2.containsKey(s2ID))
      throw new UnsupportedDuplicateException("system2 id ["+s2ID+"] already in the map!");
    
    mapRGold.put(goldID, new Integer(goldR.getSelectedCategory()));
    int category = system1R.getSelectedCategory();
    mapRSystem1.put(s1ID, new IntPair(category, category));
    category = system2R.getSelectedCategory();
    mapRSystem2.put(s2ID, new IntPair(category, category));
  }
  
  /**
   * Estimates the significance level using stratified shuffling.
   * <pre>
   * Common used significance levels: 
   *    .1  (90% confidence level) 
   *    .05 (95% confidence level) 
   *    .01 (99% confidence level) 
   * </pre>
   */
  public void estimateSignificanceLevel() {
    Random rnd = new Random(new Date().getTime());
    double origDiff = 0.0;
    double crtDiff = 0.0;
    
    DoublePair acc2 = computeAccuracy();
    origDiff = Math.abs(acc2.getFirst() - acc2.getSecond());    
    out.printf("system_1 acc:[%2.4f]  system_2 acc:[%2.4f] diff[%2.4f]\n", acc2.getFirst(), acc2.getSecond(), origDiff);
    
    int nge = 0;
    for(int s = 1; s <= NUMBER_OF_SHUFFLES; s++) {
      if(s % 50 == 0) out.print('.');
      for(Enumeration<Integer> keys = mapRGold.keys(); keys.hasMoreElements();) {
        int key = keys.nextElement();
        if(rnd.nextDouble() < 0.5)
          swap(mapRSystem1.get(key), mapRSystem2.get(key));
        else 
          dup(mapRSystem1.get(key), mapRSystem2.get(key));
      }
      acc2 = computeAccuracy();
      crtDiff = Math.abs(acc2.getFirst() - acc2.getSecond());
      if(crtDiff >= origDiff) nge++;
    }
    out.println();
		double p = (double)(nge+1) / (NUMBER_OF_SHUFFLES + 1);
    out.printf("nge[%1d] p-value:%1.5f\n",nge,p);
  }
  
  /**
   * Computes the accuracy measures for the two systems
   *
   * @returns the accuracy of the two systems
   */
  private DoublePair computeAccuracy() {
    int c1 = 0;
    int c2 = 0;
    int t = 0;
    
    for(Enumeration<Integer> keys = mapRGold.keys(); keys.hasMoreElements();) {
      int key = keys.nextElement();        
      int goldR = mapRGold.get(key);
      if(goldR == mapRSystem1.get(key).getSecond()) c1++;
      if(goldR == mapRSystem2.get(key).getSecond()) c2++;
      t++;        
    }    
    return new DoublePair((double)c1/t, (double)c2/t);
  }

  /**
   * Swaps the first values of two specified pairs and copies them into their second values
   * 
   */
  private void swap(IntPair p1, IntPair p2) {
    p1.setSecond(p2.getFirst());
    p2.setSecond(p1.getFirst());   
  }

  /**
   * Copies the first values of two specified pairs into their second values
   *
   */
  private void dup(IntPair p1, IntPair p2) {
    p1.setSecond(p1.getFirst());
    p2.setSecond(p2.getFirst());    
  }  

  
  public static void main(String[] args) {
    if(args.length != 3) {
      out.println("\n**** Error: Wrong number of arguments ["+args.length+"]**** \n");
      out.println("The COPA evaluator needs to receive three input arguments:");
      out.println(" 1) the file with gold responses");
      out.println(" 2) the file with the responses of the first system");
      out.println(" 3) the file with the responses of the second system");
      System.exit(0);
    }
    
    File f1 = new File(args[0]);
    File f2 = new File(args[1]);
    File f3 = new File(args[2]);
    
    if(!f1.isFile() || !f2.isFile() || !f3.isFile())
      throw new IllegalArgumentException("There is at least on argument that is not a valid file path! ["+args[0]+"] ["+args[1]+"] ["+args[2]+"]");
    
    
    SignificanceTest test = new SignificanceTest();
    test.loadResponses(args[0], args[1], args[2]);
    test.estimateSignificanceLevel();        
  }
}
